@@ -12,7 +12,7 @@ from __future__ import absolute_import
 #
 # Imports
 #
-
+import copy
 import errno
 import itertools
 import os
@@ -26,10 +26,12 @@ import warnings
 from collections import deque
 from functools import partial
 
-from . import Event, Process, cpu_count
+from . import cpu_count, get_context
 from . import util
-from .common import pickle_loads, reset_signals, restart_state
-from .compat import get_errno, send_offset
+from .common import (
+    TERM_SIGNAL, human_status, pickle_loads, reset_signals, restart_state,
+)
+from .compat import get_errno, mem_rss, send_offset
 from .einfo import ExceptionInfo
 from .dummy import DummyProcess
 from .exceptions import (
@@ -42,7 +44,11 @@ from .exceptions import (
     WorkerLostError,
 )
 from .five import Empty, Queue, range, values, reraise, monotonic
-from .util import Finalize, debug
+from .util import Finalize, debug, warning
+
+MAXMEM_USED_FMT = """\
+child process exiting after exceeding memory limit ({0}KiB / {1}KiB)
+"""
 
 PY3 = sys.version_info[0] == 3
 
@@ -51,7 +57,7 @@ if platform.system() == 'Windows':  # pragma: no cover
     # handled by # any process, so this is needed to terminate the task
     # *and its children* (if any).
     from ._win import kill_processtree as _kill  # noqa
-    SIGKILL = signal.SIGTERM
+    SIGKILL = TERM_SIGNAL
 else:
     from os import kill as _kill                 # noqa
     SIGKILL = signal.SIGKILL
@@ -69,10 +75,6 @@ else:
     # Semaphore is a factory function pointing to _Semaphore
     _Semaphore = threading._Semaphore  # noqa
 
-SIGMAP = dict(
-    (getattr(signal, n), n) for n in dir(signal) if n.startswith('SIG')
-)
-
 #
 # Constants representing the state of a pool
 #
@@ -108,6 +110,8 @@ SIG_SOFT_TIMEOUT = getattr(signal, "SIGUSR1", None)
 
 LOST_WORKER_TIMEOUT = 10.0
 EX_OK = getattr(os, "EX_OK", 0)
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT = 300
+GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL = 0.1
 
 job_counter = itertools.count()
 
@@ -124,15 +128,6 @@ def _get_send_offset(connection):
     return native
 
 
-def human_status(status):
-    if (status or 0) < 0:
-        try:
-            return 'signal {0} ({1})'.format(-status, SIGMAP[-status])
-        except KeyError:
-            return 'signal {0}'.format(-status)
-    return 'exitcode {0}'.format(status)
-
-
 def mapstar(args):
     return list(map(*args))
 
@@ -142,12 +137,11 @@ def starmapstar(args):
 
 
 def error(msg, *args, **kwargs):
-    if util._logger:
-        util._logger.error(msg, *args, **kwargs)
+    util.get_logger().error(msg, *args, **kwargs)
 
 
 def stop_if_not_current(thread, timeout=None):
-    if thread is not threading.currentThread():
+    if thread is not threading.current_thread():
         thread.stop(timeout)
 
 
@@ -155,19 +149,22 @@ class LaxBoundedSemaphore(_Semaphore):
     """Semaphore that checks that # release is <= # acquires,
     but ignores if # releases >= value."""
 
-    def __init__(self, value=1, verbose=None):
-        if PY3:
-            _Semaphore.__init__(self, value)
-        else:
-            _Semaphore.__init__(self, value, verbose)
-        self._initial_value = value
-
     def shrink(self):
         self._initial_value -= 1
         self.acquire()
 
     if PY3:
 
+        def __init__(self, value=1, verbose=None):
+            _Semaphore.__init__(self, value)
+            self._initial_value = value
+
+        def grow(self):
+            with self._cond:
+                self._initial_value += 1
+                self._value += 1
+                self._cond.notify()
+
         def release(self):
             cond = self._cond
             with cond:
@@ -178,15 +175,12 @@ class LaxBoundedSemaphore(_Semaphore):
         def clear(self):
             while self._value < self._initial_value:
                 _Semaphore.release(self)
-
-        def grow(self):
-            with self._cond:
-                self._initial_value += 1
-                self._value += 1
-                self._cond.notify()
-
     else:
 
+        def __init__(self, value=1, verbose=None):
+            _Semaphore.__init__(self, value, verbose)
+            self._initial_value = value
+
         def grow(self):
             cond = self._Semaphore__cond
             with cond:
@@ -220,7 +214,7 @@ class MaybeEncodingError(Exception):
         super(MaybeEncodingError, self).__init__(self.exc, self.value)
 
     def __repr__(self):
-        return "<MaybeEncodingError: %s>" % str(self)
+        return "<%s: %s>" % (self.__class__.__name__, str(self))
 
     def __str__(self):
         return "Error sending result: '%r'. Reason: '%r'." % (
@@ -239,51 +233,54 @@ def soft_timeout_sighandler(signum, frame):
 #
 
 
-class Worker(Process):
-    _controlled_termination = False
-    _job_terminated = False
+class Worker(object):
 
     def __init__(self, inq, outq, synq=None, initializer=None, initargs=(),
                  maxtasks=None, sentinel=None, on_exit=None,
-                 sigprotection=True):
+                 sigprotection=True, wrap_exception=True,
+                 max_memory_per_child=None, on_ready_counter=None):
         assert maxtasks is None or (type(maxtasks) == int and maxtasks > 0)
         self.initializer = initializer
         self.initargs = initargs
         self.maxtasks = maxtasks
+        self.max_memory_per_child = max_memory_per_child
         self._shutdown = sentinel
         self.on_exit = on_exit
         self.sigprotection = sigprotection
         self.inq, self.outq, self.synq = inq, outq, synq
-        self._make_shortcuts()
+        self.wrap_exception = wrap_exception  # XXX cannot disable yet
+        self.on_ready_counter = on_ready_counter
+        self.contribute_to_object(self)
 
-        super(Worker, self).__init__()
+    def contribute_to_object(self, obj):
+        obj.inq, obj.outq, obj.synq = self.inq, self.outq, self.synq
+        obj.inqW_fd = self.inq._writer.fileno()    # inqueue write fd
+        obj.outqR_fd = self.outq._reader.fileno()  # outqueue read fd
+        if self.synq:
+            obj.synqR_fd = self.synq._reader.fileno()  # synqueue read fd
+            obj.synqW_fd = self.synq._writer.fileno()  # synqueue write fd
+            obj.send_syn_offset = _get_send_offset(self.synq._writer)
+        else:
+            obj.synqR_fd = obj.synqW_fd = obj._send_syn_offset = None
+        obj._quick_put = self.inq._writer.send
+        obj._quick_get = self.outq._reader.recv
+        obj.send_job_offset = _get_send_offset(self.inq._writer)
+        return obj
 
     def __reduce__(self):
         return self.__class__, (
             self.inq, self.outq, self.synq, self.initializer,
-            self.initargs, self.maxtasks, self._shutdown,
+            self.initargs, self.maxtasks, self._shutdown, self.on_exit,
+            self.sigprotection, self.wrap_exception, self.max_memory_per_child,
         )
 
-    def _make_shortcuts(self):
-        self.inqW_fd = self.inq._writer.fileno()    # inqueue write fd
-        self.outqR_fd = self.outq._reader.fileno()  # outqueue read fd
-        if self.synq:
-            self.synqR_fd = self.synq._reader.fileno()  # synqueue read fd
-            self.synqW_fd = self.synq._writer.fileno()  # synqueue write fd
-            self.send_syn_offset = _get_send_offset(self.synq._writer)
-        else:
-            self.synqR_fd = self.synqW_fd = self._send_syn_offset = None
-        self._quick_put = self.inq._writer.send
-        self._quick_get = self.outq._reader.recv
-        self.send_job_offset = _get_send_offset(self.inq._writer)
-
-    def run(self):
+    def __call__(self):
         _exit = sys.exit
         _exitcode = [None]
 
         def exit(status=None):
             _exitcode[0] = status
-            return _exit()
+            return _exit(status)
         sys.exit = exit
 
         pid = os.getpid()
@@ -318,10 +315,6 @@ class Worker(Process):
     def on_loop_start(self, pid):
         pass
 
-    def terminate_controlled(self):
-        self._controlled_termination = True
-        self.terminate()
-
     def prepare_result(self, result):
         return result
 
@@ -331,6 +324,7 @@ class Worker(Process):
         inqW_fd = self.inqW_fd
         synqW_fd = self.synqW_fd
         maxtasks = self.maxtasks
+        max_memory_per_child = self.max_memory_per_child or 0
         prepare_result = self.prepare_result
 
         wait_for_job = self.wait_for_job
@@ -352,38 +346,70 @@ class Worker(Process):
                 i += 1
 
         completed = 0
-        while maxtasks is None or (maxtasks and completed < maxtasks):
-            req = wait_for_job()
-            if req:
-                type_, args_ = req
-                assert type_ == TASK
-                job, i, fun, args, kwargs = args_
-                put((ACK, (job, i, now(), pid, synqW_fd)))
-                if _wait_for_syn:
-                    confirm = wait_for_syn(job)
-                    if not confirm:
-                        continue  # received NACK
-                try:
-                    result = (True, prepare_result(fun(*args, **kwargs)))
-                except Exception:
-                    result = (False, ExceptionInfo())
-                try:
-                    put((READY, (job, i, result, inqW_fd)))
-                except Exception as exc:
-                    _, _, tb = sys.exc_info()
+        try:
+            while maxtasks is None or (maxtasks and completed < maxtasks):
+                req = wait_for_job()
+                if req:
+                    type_, args_ = req
+                    assert type_ == TASK
+                    job, i, fun, args, kwargs = args_
+                    put((ACK, (job, i, now(), pid, synqW_fd)))
+                    if _wait_for_syn:
+                        confirm = wait_for_syn(job)
+                        if not confirm:
+                            continue  # received NACK
                     try:
-                        wrapped = MaybeEncodingError(exc, result[1])
-                        einfo = ExceptionInfo((
-                            MaybeEncodingError, wrapped, tb,
-                        ))
-                        put((READY, (job, i, (False, einfo), inqW_fd)))
-                    finally:
-                        del(tb)
-                completed += 1
-        debug('worker exiting after %d tasks', completed)
-        if maxtasks:
-            return EX_RECYCLE if completed == maxtasks else EX_FAILURE
-        return EX_OK
+                        result = (True, prepare_result(fun(*args, **kwargs)))
+                    except Exception:
+                        result = (False, ExceptionInfo())
+                    try:
+                        put((READY, (job, i, result, inqW_fd)))
+                    except Exception as exc:
+                        _, _, tb = sys.exc_info()
+                        try:
+                            wrapped = MaybeEncodingError(exc, result[1])
+                            einfo = ExceptionInfo((
+                                MaybeEncodingError, wrapped, tb,
+                            ))
+                            put((READY, (job, i, (False, einfo), inqW_fd)))
+                        finally:
+                            del(tb)
+                    completed += 1
+                    if max_memory_per_child > 0:
+                        used_kb = mem_rss()
+                        if used_kb <= 0:
+                            error('worker unable to determine memory usage')
+                        if used_kb > 0 and used_kb > max_memory_per_child:
+                            warning(MAXMEM_USED_FMT.format(
+                                used_kb, max_memory_per_child))
+                            return EX_RECYCLE
+
+            debug('worker exiting after %d tasks', completed)
+            if maxtasks:
+                return EX_RECYCLE if completed == maxtasks else EX_FAILURE
+            return EX_OK
+        finally:
+            # Before exiting the worker, we want to ensure that that all
+            # messages produced by the worker have been consumed by the main
+            # process. This prevents the worker being terminated prematurely
+            # and messages being lost.
+            self._ensure_messages_consumed(completed=completed)
+
+    def _ensure_messages_consumed(self, completed):
+        """ Returns true if all messages sent out have been received and
+        consumed within a reasonable amount of time """
+
+        if not self.on_ready_counter:
+            return False
+
+        for retry in range(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_LIMIT):
+            if self.on_ready_counter.value >= completed:
+                debug('ensured messages consumed after %d retries', retry)
+                return True
+            time.sleep(GUARANTEE_MESSAGE_CONSUMPTION_RETRY_INTERVAL)
+        warning('could not ensure all messages were consumed prior to '
+                'exiting')
+        return False
 
     def after_fork(self):
         if hasattr(self.inq, '_writer'):
@@ -479,7 +505,7 @@ class PoolThread(DummyProcess):
         except RestartFreqExceeded as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
                   exc_info=1)
-            _kill(os.getpid(), signal.SIGTERM)
+            _kill(os.getpid(), TERM_SIGNAL)
             sys.exit()
         except Exception as exc:
             error("Thread %r crashed: %r", type(self).__name__, exc,
@@ -545,20 +571,23 @@ class Supervisor(PoolThread):
 
 class TaskHandler(PoolThread):
 
-    def __init__(self, taskqueue, put, outqueue, pool):
+    def __init__(self, taskqueue, put, outqueue, pool, cache):
         self.taskqueue = taskqueue
         self.put = put
         self.outqueue = outqueue
         self.pool = pool
+        self.cache = cache
         super(TaskHandler, self).__init__()
 
     def body(self):
+        cache = self.cache
         taskqueue = self.taskqueue
         put = self.put
 
         for taskseq, set_length in iter(taskqueue.get, None):
+            task = None
+            i = -1
             try:
-                i = -1
                 for i, task in enumerate(taskseq):
                     if self._state:
                         debug('task handler found thread._state != RUN')
@@ -568,15 +597,25 @@ class TaskHandler(PoolThread):
                     except IOError:
                         debug('could not put task on queue')
                         break
+                    except Exception:
+                        job, ind = task[:2]
+                        try:
+                            cache[job]._set(ind, (False, ExceptionInfo()))
+                        except KeyError:
+                            pass
                 else:
                     if set_length:
                         debug('doing set_length()')
                         set_length(i + 1)
                     continue
                 break
-            except Exception as exc:
-                error('Task Handler ERROR: %r', exc, exc_info=1)
-                break
+            except Exception:
+                job, ind = task[:2] if task else (0, 0)
+                if job in cache:
+                    cache[job]._set(ind + 1, (False, ExceptionInfo()))
+                if set_length:
+                    util.debug('doing set_length()')
+                    set_length(i + 1)
         else:
             debug('task handler got sentinel')
 
@@ -660,7 +699,11 @@ class TimeoutHandler(PoolThread):
     def _trywaitkill(self, worker):
         debug('timeout: sending TERM to %s', worker._name)
         try:
-            worker.terminate()
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGTERM) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGTERM)
+            else:
+                worker.terminate()
         except OSError:
             pass
         else:
@@ -668,12 +711,15 @@ class TimeoutHandler(PoolThread):
                 return
         debug('timeout: TERM timed-out, now sending KILL to %s', worker._name)
         try:
-            _kill(worker.pid, SIGKILL)
+            if os.getpgid(worker.pid) == worker.pid:
+                debug("worker %s is a group leader. It is safe to kill (SIGKILL) the whole group", worker.pid)
+                os.killpg(os.getpgid(worker.pid), signal.SIGKILL)
+            else:
+                _kill(worker.pid, SIGKILL)
         except OSError:
             pass
 
     def handle_timeouts(self):
-        cache = self.cache
         t_hard, t_soft = self.t_hard, self.t_soft
         dirty = set()
         on_soft_timeout = self.on_soft_timeout
@@ -687,12 +733,16 @@ class TimeoutHandler(PoolThread):
 
         # Inner-loop
         while self._state == RUN:
+            # Perform a shallow copy before iteration because keys can change.
+            # A deep copy fails (on shutdown) due to thread.lock objects.
+            # https://github.com/celery/billiard/issues/260
+            cache = copy.copy(self.cache)
 
             # Remove dirty items not in cache anymore
             if dirty:
                 dirty = set(k for k in dirty if k in cache)
 
-            for i, job in list(cache.items()):
+            for i, job in cache.items():
                 ack_time = job._time_accepted
                 soft_timeout = job._soft_timeout
                 if soft_timeout is None:
@@ -729,7 +779,7 @@ class ResultHandler(PoolThread):
 
     def __init__(self, outqueue, get, cache, poll,
                  join_exited_workers, putlock, restart_state,
-                 check_timeouts, on_job_ready):
+                 check_timeouts, on_job_ready, on_ready_counters=None):
         self.outqueue = outqueue
         self.get = get
         self.cache = cache
@@ -741,6 +791,7 @@ class ResultHandler(PoolThread):
         self._shutdown_complete = False
         self.check_timeouts = check_timeouts
         self.on_job_ready = on_job_ready
+        self.on_ready_counters = on_ready_counters
         self._make_methods()
         super(ResultHandler, self).__init__()
 
@@ -769,6 +820,14 @@ class ResultHandler(PoolThread):
                 item = cache[job]
             except KeyError:
                 return
+
+            if self.on_ready_counters:
+                worker_pid = next(iter(item.worker_pids()), None)
+                if worker_pid and worker_pid in self.on_ready_counters:
+                    on_ready_counter = self.on_ready_counters[worker_pid]
+                    with on_ready_counter.get_lock():
+                        on_ready_counter.value += 1
+
             if not item.ready():
                 if putlock is not None:
                     putlock.release()
@@ -779,7 +838,7 @@ class ResultHandler(PoolThread):
 
         def on_death(pid, exitcode):
             try:
-                os.kill(pid, signal.SIGTERM)
+                os.kill(pid, TERM_SIGNAL)
             except OSError as exc:
                 if get_errno(exc) != errno.ESRCH:
                     raise
@@ -905,6 +964,7 @@ class Pool(object):
     '''
     Class which supports an async version of applying functions to arguments.
     '''
+    _wrap_exception = True
     Worker = Worker
     Supervisor = Supervisor
     TaskHandler = TaskHandler
@@ -926,7 +986,11 @@ class Pool(object):
                  allow_restart=False,
                  synack=False,
                  on_process_exit=None,
+                 context=None,
+                 max_memory_per_child=None,
+                 enable_timeouts=False,
                  **kwargs):
+        self._ctx = context or get_context()
         self.synack = synack
         self._setup_queues()
         self._taskqueue = Queue()
@@ -935,6 +999,7 @@ class Pool(object):
         self.timeout = timeout
         self.soft_timeout = soft_timeout
         self._maxtasksperchild = maxtasksperchild
+        self._max_memory_per_child = max_memory_per_child
         self._initializer = initializer
         self._initargs = initargs
         self._on_process_exit = on_process_exit
@@ -947,6 +1012,12 @@ class Pool(object):
         self.readers = {}
         self.allow_restart = allow_restart
 
+        self.enable_timeouts = bool(
+            enable_timeouts or
+            self.timeout is not None or
+            self.soft_timeout is not None
+        )
+
         if soft_timeout and SIG_SOFT_TIMEOUT is None:
             warnings.warn(UserWarning(
                 "Soft timeouts are not supported: "
@@ -964,8 +1035,11 @@ class Pool(object):
         if on_process_exit is not None and not callable(on_process_exit):
             raise TypeError('on_process_exit must be callable')
 
+        self._Process = self._ctx.Process
+
         self._pool = []
         self._poolctrl = {}
+        self._on_ready_counters = {}
         self.putlocks = putlocks
         self._putlock = semaphore or LaxBoundedSemaphore(self._processes)
         for i in range(self._processes):
@@ -978,25 +1052,30 @@ class Pool(object):
         self._task_handler = self.TaskHandler(self._taskqueue,
                                               self._quick_put,
                                               self._outqueue,
-                                              self._pool)
+                                              self._pool,
+                                              self._cache)
         if threads:
             self._task_handler.start()
 
-        # Thread killing timedout jobs.
-        self._timeout_handler = self.TimeoutHandler(
-            self._pool, self._cache,
-            self.soft_timeout, self.timeout,
-        )
-        self._timeout_handler_mutex = Lock()
-        self._timeout_handler_started = False
-        if self.timeout is not None or self.soft_timeout is not None:
-            self._start_timeout_handler()
-
-        # If running without threads, we need to check for timeouts
-        # while waiting for unfinished work at shutdown.
         self.check_timeouts = None
-        if not threads:
-            self.check_timeouts = self._timeout_handler.handle_event
+
+        # Thread killing timedout jobs.
+        if self.enable_timeouts:
+            self._timeout_handler = self.TimeoutHandler(
+                self._pool, self._cache,
+                self.soft_timeout, self.timeout,
+            )
+            self._timeout_handler_mutex = Lock()
+            self._timeout_handler_started = False
+            self._start_timeout_handler()
+            # If running without threads, we need to check for timeouts
+            # while waiting for unfinished work at shutdown.
+            if not threads:
+                self.check_timeouts = self._timeout_handler.handle_event
+        else:
+            self._timeout_handler = None
+            self._timeout_handler_started = False
+            self._timeout_handler_mutex = None
 
         # Thread processing results in the outqueue.
         self._result_handler = self.create_result_handler()
@@ -1015,12 +1094,19 @@ class Pool(object):
             exitpriority=15,
         )
 
+    def Process(self, *args, **kwds):
+        return self._Process(*args, **kwds)
+
+    def WorkerProcess(self, worker):
+        return worker.contribute_to_object(self.Process(target=worker))
+
     def create_result_handler(self, **extra_kwargs):
         return self.ResultHandler(
             self._outqueue, self._quick_get, self._cache,
             self._poll_result, self._join_exited_workers,
             self._putlock, self.restart_state, self.check_timeouts,
-            self.on_job_ready, **extra_kwargs
+            self.on_job_ready, on_ready_counters=self._on_ready_counters,
+            **extra_kwargs
         )
 
     def on_job_ready(self, job, i, obj, inqW_fd):
@@ -1051,15 +1137,19 @@ class Pool(object):
         return self._inqueue, self._outqueue, None
 
     def _create_worker_process(self, i):
-        sentinel = Event() if self.allow_restart else None
+        sentinel = self._ctx.Event() if self.allow_restart else None
         inq, outq, synq = self.get_process_queues()
-        w = self.Worker(
+        on_ready_counter = self._ctx.Value('i')
+        w = self.WorkerProcess(self.Worker(
             inq, outq, synq, self._initializer, self._initargs,
             self._maxtasksperchild, sentinel, self._on_process_exit,
             # Need to handle all signals if using the ipc semaphore,
             # to make sure the semaphore is released.
             sigprotection=self.threads,
-        )
+            wrap_exception=self._wrap_exception,
+            max_memory_per_child=self._max_memory_per_child,
+            on_ready_counter=on_ready_counter,
+        ))
         self._pool.append(w)
         self._process_register_queues(w, (inq, outq, synq))
         w.name = w.name.replace('Process', 'PoolWorker')
@@ -1067,6 +1157,7 @@ class Pool(object):
         w.index = i
         w.start()
         self._poolctrl[w.pid] = sentinel
+        self._on_ready_counters[w.pid] = on_ready_counter
         if self.on_process_up:
             self.on_process_up(w)
         return w
@@ -1117,6 +1208,7 @@ class Pool(object):
                 self.process_flush_queues(worker)
                 del self._pool[i]
                 del self._poolctrl[worker.pid]
+                del self._on_ready_counters[worker.pid]
         if cleaned:
             all_pids = [w.pid for w in self._pool]
             for job in list(self._cache.values()):
@@ -1214,7 +1306,6 @@ class Pool(object):
         for worker in self._pool:
             if not self._worker_active(worker):
                 yield worker
-        raise StopIteration()
 
     def _worker_active(self, worker):
         for job in values(self._cache):
@@ -1270,9 +1361,8 @@ class Pool(object):
                 raise
 
     def _setup_queues(self):
-        from billiard.queues import SimpleQueue
-        self._inqueue = SimpleQueue()
-        self._outqueue = SimpleQueue()
+        self._inqueue = self._ctx.SimpleQueue()
+        self._outqueue = self._ctx.SimpleQueue()
         self._quick_put = self._inqueue._writer.send
         self._quick_get = self._outqueue._reader.recv
 
@@ -1285,7 +1375,7 @@ class Pool(object):
     def _start_timeout_handler(self):
         # ensure more than one thread does not start the timeout handler
         # thread at once.
-        if self.threads:
+        if self.threads and self._timeout_handler is not None:
             with self._timeout_handler_mutex:
                 if not self._timeout_handler_started:
                     self._timeout_handler_started = True
@@ -1447,7 +1537,7 @@ class Pool(object):
         proc, _ = self._process_by_pid(pid)
         if proc is not None:
             try:
-                _kill(pid, sig or signal.SIGTERM)
+                _kill(pid, sig or TERM_SIGNAL)
             except OSError as exc:
                 if get_errno(exc) != errno.ESRCH:
                     raise
@@ -1532,7 +1622,7 @@ class Pool(object):
         stop_if_not_current(self._result_handler)
         debug('result handler joined')
         for i, p in enumerate(self._pool):
-            debug('joining worker %s/%s (%r)', i+1, len(self._pool), p)
+            debug('joining worker %s/%s (%r)', i + 1, len(self._pool), p)
             if p._popen is not None:  # process started?
                 p.join()
         debug('pool join complete')
@@ -1604,6 +1694,9 @@ class Pool(object):
                         p.join()
             debug('pool workers joined')
 
+        inqueue.close()
+        outqueue.close()
+
     @property
     def process_sentinels(self):
         return [w._popen.sentinel for w in self._pool]
@@ -1649,7 +1742,8 @@ class ApplyResult(object):
         cache[self._job] = self
 
     def __repr__(self):
-        return '<Result: {id} ack:{ack} ready:{ready}>'.format(
+        return '<%s: {id} ack:{ack} ready:{ready}>'.format(
+            self.__class__.__name__,
             id=self._job, ack=self._accepted, ready=self.ready(),
         )
 
@@ -1931,7 +2025,7 @@ class IMapUnorderedIterator(IMapIterator):
 
 class ThreadPool(Pool):
 
-    from billiard.dummy import Process as DummyProcess
+    from .dummy import Process as DummyProcess
     Process = DummyProcess
 
     def __init__(self, processes=None, initializer=None, initargs=()):
